from functools import wraps
from logging import Logger
from typing import Callable, TypeVar, Any

from elasticsearch import Elasticsearch
from tenacity import Retrying, wait_exponential, wait_random, stop_after_attempt, RetryError

T = TypeVar("T")


class Error(Exception):
    pass


class MaxRetriesExceededError(Error):
    pass


class ResourceUsageError(Error):
    pass


class ESGuard:
    def __init__(
        self,
        es: Elasticsearch = Elasticsearch(),
        os_cpu_percent: int = 90,
        os_mem_used_percent: int = 90,
        jvm_mem_heap_used_percent: int = 90,
        retry_backoff_sec: int = 1,
        max_retries: int = 3,
        logger: Logger = None,
    ) -> None:
        self.logger = logger
        self.es = es
        self.os_cpu_percent = os_cpu_percent
        self.os_mem_used_percent = os_mem_used_percent
        self.jvm_mem_heap_used_percent = jvm_mem_heap_used_percent
        self.retry_backoff_sec = retry_backoff_sec
        self.max_retries = max_retries

    def _warning(self, message: str) -> None:
        if self.logger is None:
            return
        self.logger.warning(message)

    def _wait(self):
        res = self.es.nodes.stats(metric=["os", "jvm"])
        for k in res["nodes"].keys():
            cpu = res["nodes"][k]["os"]["cpu"]["percent"]
            if cpu >= self.os_cpu_percent and self.os_cpu_percent > 0:
                self._warning(f"node({k}) OS CPU usage {cpu}% over {self.os_cpu_percent}% ")
                raise ResourceUsageError(f"node({k}) OS CPU usage {cpu}% over {self.os_cpu_percent}% ")

            mem = res["nodes"][k]["os"]["mem"]["used_percent"]
            if mem >= self.os_mem_used_percent and self.os_mem_used_percent > 0:
                self._warning(f"node({k}) OS MEM usage {mem}% over {self.os_mem_used_percent}% ")
                raise ResourceUsageError(f"node({k}) OS MEM usage {mem}% over {self.os_mem_used_percent}% ")

            jvm_heap = res["nodes"][k]["jvm"]["mem"]["heap_used_percent"]
            if (jvm_heap >= self.jvm_mem_heap_used_percent and self.jvm_mem_heap_used_percent > 0):
                self._warning(f"node({k}) JVM heap usage {jvm_heap}% over {self.jvm_mem_heap_used_percent}% ")
                raise ResourceUsageError(f"node({k}) JVM heap usage {jvm_heap}% over {self.jvm_mem_heap_used_percent}% ")

    def _get_retryer(self):
        return Retrying(wait=wait_exponential(min=self.retry_backoff_sec) + wait_random(min=0, max=1), stop=stop_after_attempt(self.max_retries))

    def decotator(self) -> Callable[[], T]:
        retryer = self._get_retryer()

        def _retry(func) -> Callable[[], T]:
            @wraps(func)
            def wrapper(*args, **kwargs) -> Any:
                try:
                    retryer(self._wait)
                except RetryError:
                    raise MaxRetriesExceededError(f"max retries exceeded {self.max_retries}")
                return func(*args, **kwargs)

            return wrapper

        return _retry



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esguard']

package_data = \
{'': ['*']}

install_requires = \
['docker>=5.0.0,<6.0.0',
 'elasticsearch>=7.12.1,<8.0.0',
 'six>=1.16.0,<2.0.0',
 'tenacity>=8.0.1,<9.0.0']

setup_kwargs = {
    'name': 'esguard',
    'version': '0.1.4',
    'description': 'esguard provides a Python decorator that waits for processing while monitoring the load of Elasticsearch.',
    'long_description': '<p align="center">\n  <img alt="esguard-logo" src="esguard.png" height="100" />\n  <h2 align="center">esguard</h2>\n  <p align="center">esguard provides a Python decorator that waits for processing while monitoring the load of Elasticsearch.</p>\n</p>\n\n[![PyPi version](https://img.shields.io/pypi/v/esguard.svg)](https://pypi.python.org/pypi/esguard/) [![](https://img.shields.io/badge/python-3.7+-blue.svg)](https://www.python.org/downloads/release/python-390/) ![PyTest](https://github.com/po3rin/esguard/workflows/PyTest/badge.svg)\n\n## Quick Start\n\nYou need to launch elasticsearch before quick start.\n\n```python\nfrom esguard import ESGuard\n\n\n@ESGuard(os_cpu_percent=95).decotator()\ndef mock_func(x):\n    return x\n\nself.assertEqual(mock_func(1), 1)\n```\n\n## Test\n\nYou need to launch elasticsearch before testing.\n\n```sh\n$ docker compose up -d --build\n$ poetry run pytest\n```\n',
    'author': 'po3rin',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/po3rin/esguard',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
